!
! Copyright (C) 2015-2016 Aihui Zhou's group
!
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!-------------------------------------------------------------------------------
!
! We propose some parallel orbital updating based plane wave basis methods
! for electronic structure calculations, which aims to the solution of the corresponding eigenvalue
! problems. Compared to the traditional plane wave methods, our methods have the feature of two level
! parallelization, which make them have great advantage in large-scale parallelization.
!
! The approach following Algorithm is the parallel orbital updating algorithm:
! 1. Choose initial $E_{\mathrm{cut}}^{(0)}$ and then obtain $V_{N_G^{0}}$, use the SCF method to solve
!    the Kohn-Sham equation in $V_{G_0}$ and get the initial $(\lambda_i^{0},u_i^{0}), i=1, \cdots, N$ 
!    and let $n=0$.
! 2. For $i=1,2,\ldots,N$, find $e_i^{n+1/2}\in V_{G_n}$ satisfying
!    $$a(\rho_{in}^{n}; e_i^{n+1/2}, v) = -[(a(\rho_{in}^{n}; u_i^{n}, v) - \lambda_i^{n} (u_i^{n}, v))]  $$
!    in parallel , where $\rho_{in}^{n}$ is the input charge density obtained by the orbits obtained in the 
!    $n$-th iteration or the former iterations.
! 3. Find $\{\lambda_i^{n+1},u_i^{n+1}\} \in \mathbf{R}\times \tilde{V}_N$   satisfying
!      $$a(\tilde{\rho}; u_i^{n+1}, v) = ( \lambda_i^{n+1}u_i^{n+1}, v) \quad  \forall v \in \tilde{V}_N$$
!      where $\tilde{V}_N = \mathrm{span}\{e_1^{n+1/2},\ldots,e_N^{n+1/2},u_1^{n},\ldots,u_N^{n}\}$, 
!      $\tilde{\rho}(x)$ is the input charge density obtained from the previous orbits.
! 4. Convergence check: if not converged, set $n=n+1$, go to step 2; else,  stop.
!
! You can see the detailed information through
!  X. Dai, X. Gong, A. Zhou, J. Zhu,
!   A parallel orbital-updating approach for electronic structure calculations, arXiv:1405.0260 (2014).
! X. Dai, Z. Liu, X. Zhang, A. Zhou,
!  A Parallel Orbital-updating Based Optimization Method for Electronic Structure Calculations, 
!   arXiv:1510.07230 (2015).
! Yan Pan, Xiaoying Dai, Xingao Gong, Stefano de Gironcoli, Gian-Marco Rignanese, and Aihui Zhou,
!  A Parallel Orbital-updating Based Plane Wave Basis Method. J. Comp. Phys. 348, 482-492 (2017).
!
! The file is written mainly by Stefano de Gironcoli and Yan Pan.
!
!-------------------------------------------------------------------------------
SUBROUTINE paro( h_psi, h_1psi, s_psi, s_1psi, hs_1psi, precondition, &
                 npwx, npw, nbnd, evc, eig, btype, ethr, notconv, nhpsi, paro_flag )
  !-------------------------------------------------------------------------------
  !paro_flag = 0: original parallel orbital-updating method
  !paro_flag = 1: modified parallel orbital-updating method

  ! global variables
  USE util_param,          ONLY : DP, stdout
  USE mp_bands_util,       ONLY : my_bgrp_id, inter_bgrp_comm
  USE mp,                  ONLY : mp_sum

  IMPLICIT NONE

  ! I/O variables
  INTEGER, INTENT(IN)        :: npw, npwx, nbnd
  COMPLEX(DP), INTENT(INOUT) :: evc(npwx,nbnd)
  REAL(DP), INTENT(IN)       :: ethr
  REAL(DP), INTENT(IN)       :: precondition(npwx)
  REAL(DP), INTENT(INOUT)    :: eig(nbnd)
  INTEGER, INTENT(IN)        :: btype(nbnd)
  INTEGER, INTENT(OUT)       :: notconv, nhpsi
  INTEGER, INTENT(IN)        :: paro_flag
  
  ! local variables (used in the call to cegterg )
  INTEGER, PARAMETER :: npol=1
  LOGICAL :: overlap
  !------------------------------------------------------------------------
  EXTERNAL h_psi, h_1psi, s_psi, s_1psi, hs_1psi
  ! subroutine h_psi  (npwx,npw,nvec,evc,hpsi)  computes H*evc  using band parallelization
  ! subroutine h_1psi (npwx,npw,evc,hpsi)       computes H*evc  for a single band
  ! subroutine s_psi  (npwx,npw,nvec,evc,spsi)  computes S*evc  using band parallelization
  ! subroutine s_1psi (npwx,npw,evc,spsi)       computes S*evc  for a single band
  ! subroutine hs_1psi(npwx,npw,evc,hpsi,spsi)  computes H*evc and S*evc for a single band

  !
  ! ... local variables
  !
  INTEGER :: iter, itry, paro_ntr
  REAL(DP), ALLOCATABLE    :: ew(:)
  COMPLEX(DP), ALLOCATABLE :: evc2(:,:)
  LOGICAL, ALLOCATABLE     :: conv(:)

  REAL(DP):: paro_shift
  INTEGER :: ibnd, ibnd_start, ibnd_end

  !
  ! ... init local variables
  !
  iter = 0
  nhpsi = 0
  paro_ntr = 20
  !
  write (6,*) ' paro_flag = ', paro_flag

  IF (paro_flag == 0) THEN ! original method
     ALLOCATE ( ew(nbnd), evc2(npwx,nbnd),   conv(nbnd) )
  ELSE IF (paro_flag == 1) THEN  ! modified method
     ALLOCATE ( ew(nbnd), evc2(npwx,2*nbnd), conv(nbnd) )
  ELSE
     WRITE(stdout,*) 'wrong setting of paro_flag!! '
  END IF

  conv(:) =  .FALSE.
     
  ParO_loop : &
  DO itry = 1,paro_ntr
     write (6,*) ' paro_itry =', itry
     paro_shift = 2.0d0 * ( MAXVAL(eig(1:nbnd)) - MINVAL(eig(1:nbnd)) )
     write (6,*) ' paro_shift =', paro_shift

     IF (paro_flag == 0) THEN ! original parallel orbital-updating 

        evc2(:,1:nbnd) = evc(:,1:nbnd)
       ! ew(:) = eig(:)
        notconv = nbnd
       ! do ibnd=1,nbnd
       !   if (.not.conv(ibnd) ) then
       !     notconv = notconv+1
       !     evc(:,notconv) = evc2(:,ibnd)
       !     eig(notconv) = ew(ibnd)
       !   end if
       ! end do
        CALL divide(inter_bgrp_comm,notconv,ibnd_start,ibnd_end)
        IF ( ibnd_start > 1  ) evc(:, 1:ibnd_start-1 ) = (0.0_dp,0.0_dp)
        DO ibnd=ibnd_start,ibnd_end
           !CALL pcg(hs_1psi, npw, npwx, evc(:,ibnd), ethr, iter, precondition, eig(ibnd), nhpsi)
           CALL pcg_origin( h_1psi, s_1psi, overlap, &
                            npw, npwx, evc(:,ibnd), ethr, iter, precondition, eig(ibnd))
        END DO
        IF ( ibnd_end < nbnd ) evc(:, ibnd_end+1:nbnd) = (0.0_dp,0.0_dp)
        CALL mp_sum(evc,inter_bgrp_comm)
        !evc2(:,nbnd+1:2*nbnd) = evc(:,1:nbnd)
        !CALL rotate_wfc ( npwx, npw, nbnd, gstart, nbnd, evc2, npol, overlap, evc, eig(:))
        ! evc2(:,:) = evc2(:,:) + evc(:,:)
        evc2(:,:) = evc(:,:)
        CALL rotate_wfc_gamma (h_psi, s_psi, overlap, npwx, npw, nbnd, nbnd, evc2, evc, eig )

        IF (my_bgrp_id==0) nhpsi = nhpsi + (nbnd+notconv)
        conv(:) = ABS(ew(:)-eig(:)) .lt. ethr ; notconv = COUNT ( .not.conv(:) )
        IF ( notconv == 0 ) EXIT ParO_loop
        notconv = 0

     ELSE IF (paro_flag == 1) THEN ! modified parallel orbital-updating method

        evc2(:,1:nbnd) = evc(:,1:nbnd) ! save evc for later use
        call  s_psi  (npwx,npw,nbnd,evc,evc2(1,nbnd+1) ) ! computes S*evc needed to compute Pv
        ew = eig
        notconv=0
        DO ibnd=1,nbnd ! pack unconverged roots
           IF (.NOT.conv(ibnd) ) THEN
              notconv = notconv+1
              evc(:,notconv) = evc(:,ibnd)
              eig(notconv) = ew(ibnd)
           END IF
        END DO
        write (*,*) itry, conv
        write (6,*) ' notconv = ', notconv, ethr

        CALL divide(inter_bgrp_comm,notconv,ibnd_start,ibnd_end)
        IF ( ibnd_start > 1  ) evc(:, 1:ibnd_start-1 ) = (0.0_dp,0.0_dp)
        DO ibnd=ibnd_start,ibnd_end
           CALL pcg(hs_1psi, evc2(1,nbnd+1), paro_shift, npw, npwx, nbnd, precondition, &
                    evc(:,ibnd), ethr, iter, eig(ibnd), nhpsi)
        END DO
        IF ( ibnd_end < nbnd ) evc(:, ibnd_end+1:nbnd) = (0.0_dp,0.0_dp)
        CALL mp_sum(evc,inter_bgrp_comm)

        evc2(:,nbnd+1:nbnd+notconv) = evc(:,1:notconv) ! overwrite spsi0 nont needed anymore 
        CALL rotate_wfc_gamma( h_psi, s_psi, overlap, npwx, npw, nbnd+notconv, nbnd, evc2, evc, eig )

        IF (my_bgrp_id==0) nhpsi = nhpsi + (nbnd+notconv)
        conv(:) = ABS(ew(:)-eig(:)) .LT. ethr ; notconv = COUNT ( .NOT.conv(:) )
        IF ( notconv == 0 ) EXIT ParO_loop
        notconv = 0
     END IF
  END DO ParO_loop
  CALL mp_sum(nhpsi,inter_bgrp_comm)

  DEALLOCATE ( ew, evc2, conv )

END SUBROUTINE paro
