! Copyright (C) 2015-2016 Aihui Zhou's group
!
! This file is distributed under the terms of the
! GNU General Public License. See the file `License'
! in the root directory of the present distribution,
! or http://www.gnu.org/copyleft/gpl.txt .
!
!
!-------------------------------------------------------------------------------
!
! We propose some parallel orbital updating based plane wave basis methods
! for electronic structure calculations, which aims to the solution of the corresponding eigenvalue
! problems. Compared to the traditional plane wave methods, our methods have the feature of two level
! parallelization, which make them have great advantage in large-scale parallelization.
!
! The approach following Algorithm is the parallel orbital updating algorithm:
! 1. Choose initial $E_{\mathrm{cut}}^{(0)}$ and then obtain $V_{N_G^{0}}$, use the SCF method to solve the Kohn-Sham equation in $V_{G_0}$ and get the initial $(\lambda_i^{0},u_i^{0}), i=1, \cdots, N$ and let $n=0$.
! 2. For $i=1,2,\ldots,N$, find $u_i^{n+1/2}\in V_{N_G^{n+1}}$ satisfying $$a(\rho_{in}^{n}; u_i^{n+1/2}, v) = ( \lambda_i^{n}u_i^{n}, v) \quad  \forall v\in V_{N_G^{n+1}}$$ in parallel , where $\rho_{in}^{n}$ is the input charge density obtained by the orbits obtained in the $n$-th iteration or the previous iterations.
! 3. Find $(\lambda_i^{n+1},u_i^{n+1}) \in \mathbb{R}\times\tilde{V}_N$   satisfying
!      $$a(\rho_{in}^{n+1/2}; u_i^{n+1}, v) = ( \lambda_i^{n+1}u_i^{n+1}, v) \quad  \forall v \in \tilde{V}_N,$$
!      where $\tilde{V}_N = \mathrm{span}\{u_1^{n+1/2},\ldots,u_N^{n+1/2}\}$,  $\rho_{in}^{n+1/2}$ is the input charge density obtained
!      from $u_i^{n+1/2}$ and $u_i^{k}$ for $k\leq n$,  $i=1, \cdots, N$.
! 4. Convergence check: if not converged, set $n=n+1$, go to step 2; else,  stop.
!
! You can see the detailed information through
!  X. Dai, X. Gong, A. Zhou, J. Zhu, A parallel orbital-updating approach for electronic structure
! calculations, arXiv:1405.0260 (2014).
! X. Dai, Z. Liu, X. Zhang, A. Zhou, A Parallel Orbital-updating Based Optimization Method
! for Electronic Structure Calculations, arXiv:1510.07230 (2015).
! YAN PAN, XIAOYING DAI, XINGAO GONG, STEFANO DE GIRONCOLI, GIAN-MARCO RIGNANESE, AND AIHUI ZHOU, A PARALLEL ORBITAL-UPDATING BASED PLANE! WAVE BASIS METHOD.
! The file is written mainly by Stefano de Gironcoli and Yan Pan.
!
! The following file is for solving the step 2 of the parallel orbital updating algorithm.
!
#define ZERO ( 0.D0, 0.D0 )
#define ONE  ( 1.D0, 0.D0 )
!
!----------------------------------------------------------------------------
SUBROUTINE pcg_origin( h_psi_, s_1psi, npw, npwx, psi, dr2, iter, precondition, overlap, e )
  !----------------------------------------------------------------------------
  !
  ! ... where H is an hermitean operator, e is a real scalar,
  ! ... S is an overlap matrix, evc is a complex vector
  ! ... (real wavefunctions with only half plane waves stored)
  !
  USE paro_param,         ONLY : DP, stdout
  ! USE control_flags, ONLY : gamma_only, paro_shift
  USE mp_bands_util,      ONLY : intra_bgrp_comm, gstart
  USE mp,            ONLY : mp_sum
  !
  IMPLICIT NONE
  !
  ! Following varibales are temporary
  LOGICAL :: gamma_only=.TRUE.
  REAL(DP) :: paro_shift = 2.0D0
  !
  INTEGER, INTENT(IN) :: npw, npwx, iter
  REAL(DP), INTENT(IN) :: dr2 ! energy threshold for convergence: root improvement is stopped,
                              ! when two consecutive estimates of the root differ by less than ethr.
  LOGICAL, INTENT(IN) :: overlap ! if .FALSE. : S|psi> not needed
  REAL(DP), INTENT(IN) :: e   ! contains the estimated roots.
  REAL (DP), INTENT(IN) :: precondition(npw)
  COMPLEX(DP),INTENT(INOUT) :: psi(npwx)
!  INTEGER, INTENT(OUT) :: cg_iter ! integer  number of iterations performed
  !
  ! ... LOCAL variables
  !
  INTEGER, PARAMETER :: maxter = 200! maximum number of iterations
  !
  COMPLEX(DP), ALLOCATABLE ::  hpsi(:), & ! the product of H and psi
                               spsi(:), & ! the product of S and psi
                               b(:), r(:), p(:), sp(:), w(:),z(:) ! additional working vetors
!  LOGICAL, ALLOCATABLE :: conv ! true if the root is converged

  REAL(DP) :: g0 ,g1 ,beta, alpha, lamma, gamma, ethr_cg
  INTEGER  :: npw2, npwx2, cg_iter, l, m
  !
  REAL(DP), EXTERNAL :: DDOT

  EXTERNAL  h_psi_, s_1psi

  !
  ! the first nvec columns contain the trial eigenvectors
  !
  CALL start_clock( 'pcg_origin' )
  !
  ALLOCATE( hpsi( npwx ),  spsi( npwx ) )
  ALLOCATE( b( npwx ), r( npwx ), p( npwx ),  sp( npwx ), w( npwx ), z( npwx ) )

  lamma = paro_shift
 ! lamma = 2
  !
  npw2  = 2*npw
  npwx2  = 2*npwx
  !
  IF (iter > 200) THEN
     ethr_cg = 1.D-8
  ELSE IF (iter > 80) THEN
     ethr_cg = 1.D-6
  ELSE IF (iter > 50) THEN
     ethr_cg = 1.D-4
  ELSE
     ethr_cg = 1.D-2
  END IF
   !ethr_cg = 1.D-20
  !
  ! ... set Im[ psi(G=0) ] -  needed for numerical stability
  IF (gamma_only) THEN
     IF ( gstart == 2 ) psi(1) = CMPLX( DBLE( psi(1) ), 0.D0 ,kind=DP)
  END IF
  !
  CALL h_psi_( npwx, npw, 1, psi, hpsi ) ! apply H to a single wavefunction (no bgrp parallelization here!)
 ! hpsi = hpsi + lamma * psi              ! add an eigenvalue shift
  spsi = psi
  IF ( overlap )   CALL s_1psi( npwx, npw,  psi, spsi )
  hpsi = hpsi + lamma * spsi              ! I think it should rather be this !

  ! CG algorithm
  b(:) = e * spsi(:)
!  b(:) = lamma * psi(:) + b(:)
  b(:) = lamma * spsi(:) + b(:)
  r(:) = b(:) - hpsi(:)
  z(:) = r(:) / precondition(:)
  g0 =  DDOT( npw2, z ,1 ,r ,1)
  IF ( gamma_only ) THEN
     g0 = 2.d0 * g0 ; IF ( gstart == 2 ) g0 = g0 - CONJG (z(1)) * r(1)
  END IF
  CALL mp_sum( g0, intra_bgrp_comm )

  iterate: &
  DO cg_iter = 1, maxter

     IF ( cg_iter==1 ) THEN
        p(:) = z(:)
     ELSE
        beta = g1/g0
        g0 = g1
        p(:) = z(:) + beta * p(:)
     END IF

     CALL h_psi_( npwx, npw, 1, p, w ) ! apply H to a single wavefunction (no bgrp parallelization here!)
     sp = p ; IF ( overlap )   CALL s_1psi( npwx, npw,  p, sp )
     w = w  + lamma * sp              ! add an eigenvalue shift

     gamma = ddot( npw2 ,p ,1 ,w ,1)
     IF ( gamma_only ) THEN
         gamma = 2.d0 * gamma ; IF ( gstart == 2 ) gamma = gamma - CONJG(p(1)) * w(1)
     END IF
     CALL mp_sum( gamma, intra_bgrp_comm )
     alpha = g0/gamma

     psi(:) = psi(:) + alpha * p(:)
     r(:) = r(:) - alpha * w(:)
     z(:)   = r(:) / precondition(:)
     g1 =  ddot( npw2, z, 1, r ,1)
     IF ( gamma_only ) THEN
         g1 = 2.d0 * g1 ; IF ( gstart == 2 ) g1 = g1 - CONJG(z(1)) * r(1)
     END IF
     CALL mp_sum( g1, intra_bgrp_comm )

     IF ( ABS ( g1 ) < ethr_cg ) EXIT iterate
  END DO iterate
! IF ( ABS ( g1 ) > ethr_cg ) then
 ! write(*,*) 'g1=',g1
  !write(*,*) 'cg_iter=',cg_iter
! END IF
  !
  DEALLOCATE( b, r, p, sp, w, z )
  DEALLOCATE( hpsi,  spsi )
  !
  CALL stop_clock( 'pcg_origin' )
  !
  RETURN
  !
END SUBROUTINE pcg_origin
